// Copyright 2024 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Assembly reference of newcoro is not allowed.

package main

func main() {
	newcoro()
}

func newcoro()
